#include <stdio.h>
#include <stdlib.h>
#include "list.h"
#include "tags.h"

int feature_count(short st)
{
  int x=0;
  if (st & M_B) x++;
  if (st & M_I) x++;
  if (st & (M_EM | M_S)) x++;
  if (st & M_TT) x++;
  x+=(st&M_U)>>8;
  return x;
}

TextList* make_PL_tag(int exit)
{
  TextList *tl;
  int x=4;
  if (exit) x=5;
  tl=(TextList*)malloc(sizeof(TextList_dummy)+x);
  tl->len=x;
  x=0;
  tl->data[x++]='<';
  if (exit) tl->data[x++]='/';
  tl->data[x++]='P';
  tl->data[x++]='L';
  tl->data[x++]='>';
  return tl;
}

void fix_change(BlockNode *start)
{
  if (start->prev==NULL)  //starting node
    start->change=start->state;
  else
    start->change=start->state ^ start->prev->state;
  start->next->change=start->next->state ^ start->state;
}

int collapse_PL(Block* bl)
{
  BlockNode *c, *n;
  int count=0;
  int PL_cost, nor_cost;
  // Begin
  if (bl->start == bl->end)
    return 0;	// termination case
  c=bl->start;
  while (1)
    {
      n=c->next;
      PL_cost=2+2*feature_count(c->state);
      nor_cost=feature_count(c->state ^ c->change);
      nor_cost+=feature_count(n->state);
      nor_cost-=feature_count(c->state ^ c->change ^ n->state);
      if (PL_cost<nor_cost)
	{
	  //Write the special tags for this block
	  tl_prepend(c, make_tag_entry(c->state & 0x7F00));
	  // Do the append
	  tl_append(c, make_tag_exit(c->state & 0x7F00));
	  tl_prepend(c, make_PL_tag(0));
	  tl_append(c, make_PL_tag(1));
	  //Adjust the block features
	  if ((c->state &0xFF)==(n->state &0xFF))
	    c->state=(c->state & 0xFF)|((n->state)&0x7F00);
	  else
	    c->state=(c->state & 0xFF)|((c->change^c->state)&0x7F00);
	  fix_change(c);
	  count+=1;
	}
      c=n;
      if (c==bl->end)
	return count;
    }
}

int collapse_redun(Block* bl)
{
  BlockNode *c, *n;
  int count=0;
  if (bl->start == bl->end)
    return 0;	// termination case
  c=bl->start;
  while(1)
    {
      n=c->next;
      if(c->state == n->state)
	{
	  count+=1;
	  // Copy the text
	  c->text_end->next=n->text;
	  c->text_end=n->text_end;
	  // Move pointers to bypass the node
	  if (n->next)
	    n->next->prev=c;
	  c->next=n->next;
	  free(n);
	  if (bl->end == n)
	    {
	      bl->end=c;
	      return count;
	    }
	}
      else
	c=n;
      if (c == bl->end)
	return count;
    }
}

int collapse_single(Block* bl)
{
  BlockNode *c, *n;
  int count=0;
  short test, t1, t2;
  if (bl->start == bl->end)
    return 0;	// termination case
  c=bl->start;
  while(1)
    {
      n=c->next;
      test=c->change & n->change & c->state & 0x7C00;
      // Underline handling
      t1=(c->state & M_U)-(n->state & M_U);
      t2=(c->state & M_U)-((c->state^c->change)&M_U);
      if (t2<t1) t1=t2;  //get smallest
      if (t1>0)
	{
	  if (t1 & (~M_U))
	    fprintf(stderr,"ERROR: t1 has other bits set :%04x\n",t1);
	  test|=t1;
	}
/*       if ( ((c->state & M_U)>(n->state & M_U)) && */
/* 	   (((c->state^c->change)&M_U)<(c->state))) */
	
/*       if ( ((c->change & M_U)==(n->change & M_U))&& */
/* 	   ((c->state & M_U)>(n->state & M_U))) */
/* 	test|=(c->state & M_U)-(n->state & M_U); */
      // Size
      if (((c->change & M_SIZE)==(n->change & M_SIZE))&&(c->change & M_SIZE))
	test|=(c->state & M_SIZE);
      // Color
      if (((c->change & M_COL)==(n->change & M_COL))&&(c->change & M_COL))
	test|=(c->state & M_COL);
      if (test)
	{
	  //Write the special tags for this block
	  tl_prepend(c, make_tag_entry(test));
	  // Do the append
	  tl_append(c, make_tag_exit(test));
	  //Adjust the block features
	  c->state=c->state & (~(test & 0x7C00));
	  if (test & M_S)
	    c->state=(c->state & (~M_EM))|(n->state & M_EM);
	  if (test & M_U)
	    c->state-=(test & M_U);
	  if (test & M_SIZE)
	    c->state=(c->state & (~M_SIZE))|(n->state & M_SIZE);
	  if (test & M_COL)
	    c->state=(c->state & (~M_COL))|(n->state & M_COL);
	  fix_change(c);
	  count+=1;	  
	}
      c=n;
      if (c==bl->end)
	return count;
    }
}

int collapse_cs(Block* bl)
{
  BlockNode *c, *n;
  int count=0;
  short test;
  if (bl->start == bl->end)
    return 0;	// termination case
  c=bl->start;
  while(1)
    {
      n=c->next;
      test=0;
      // Size
      if ((c->change & M_SIZE)&&(n->change & M_SIZE))
	test|=(c->state & M_SIZE);
      // Color
      if ((c->change & M_COL)&&(n->change & M_COL))
	test|=(c->state & M_COL);
      if (test)
	{
/* 	  data_debug(bl); */
	  //Write the special tags for this block
	  tl_prepend(c, make_tag_entry(test));
	  // Do the append
	  tl_append(c, make_tag_exit(test));
	  //Adjust the block features
/* 	  c->state=c->state & (~(test & 0x7C00)); */
	  if (test & M_SIZE)
	    {
	      if (!(n->state & M_SIZE))
		c->state=c->state & (~M_SIZE);
	      else
		c->state=(c->state & (~M_SIZE))|((c->state^c->change)&M_SIZE);
	    }
/* 	    c->state=c->state & (~M_SIZE); */
	  if (test & M_COL)
	    {
	      if (!(n->state & M_COL))
		c->state=c->state & (~M_COL);
	      else
		c->state=(c->state & (~M_COL))|((c->state^c->change)&M_COL);
	    }
/* 	    c->state=c->state & (~M_COL); */
	  fix_change(c);
	  count+=1;	  
	}
      c=n;
      if (c==bl->end)
	return count;
    }
}
