#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "list.h"

#define TL_Alloc(n) (TextList*)malloc(sizeof(TextList_dummy)+n)
#define BN_Alloc() (BlockNode*)malloc(sizeof(BlockNode))
#define B_Alloc() (Block*)malloc(sizeof(Block))

// Loads everything into memory
Block* data_load(FILE* in)
{
  Block *B_top=B_Alloc();
  BlockNode *curr=BN_Alloc(), *next;
  TextList *temp;
  char c, inbuf[INBUF];
  int inlen=0;
  short new_state;
  //Data init
  B_top->start=curr;
  curr->prev=NULL;
  curr->state=0;
  curr->change=0;
  //Input Loop
  while (1)
    {
      c=getc(in);
      if (feof(in))
	break;
      if (c & 0x80)
	{	// if the state changes
	  new_state=(c<<8)&0x7F00;
	  new_state|=getc(in)&0xFF;
	  temp=TL_Alloc(inlen);
	  temp->next=NULL;
	  temp->len=inlen;
	  memcpy(temp->data,inbuf,inlen);
	  inlen=0;
	  curr->text=temp;
	  curr->text_end=temp;
	  next=BN_Alloc();
	  curr->next=next;
	  next->prev=curr;
	  next->state=new_state;
	  next->change=new_state^curr->state;
	  curr=next;
	}
      else
	{
	  inbuf[inlen++]=c;
	  if (inlen==INBUF)	// if buffer is full
	    {
	      temp=TL_Alloc(inlen);
	      temp->next=NULL;
	      temp->len=inlen;
	      memcpy(temp->data,inbuf,inlen);
	      inlen=0;
	      curr->text=temp;
	      curr->text_end=temp;
	      next=BN_Alloc();
	      curr->next=next;
	      next->prev=curr;
	      next->state=curr->state;
	      next->change=0;
	      curr=next;
	    }
	}
    }
  // Done reading the file, now to close everything...
  fclose(in);
  temp=TL_Alloc(inlen);
  temp->next=NULL;
  temp->len=inlen;
  memcpy(temp->data,inbuf,inlen);
  curr->text=temp;
  curr->text_end=temp;
  //Add one more block node to make sure
  // we get back to state 0
  next=BN_Alloc();
  temp=TL_Alloc(0);
  temp->len=0;
  temp->next=NULL;
  curr->next=next;
  next->next=NULL;
  next->prev=curr;
  next->text=temp;
  next->text_end=temp;
  next->state=0;
  next->change=curr->state;
  // Finish up
  B_top->end=next;
  return B_top;
}

// Simple function to write the data
int data_write(FILE* out, Block* d)
{
  BlockNode *bn=d->start;
  TextList *tl;
  int bc=1, x;
  while(1)
    {
      tl=bn->text;
      while(tl)
	{
	  //printf("%d ", tl->len);
	  for (x=0;x<(tl->len);x++)
	    putc(tl->data[x],out);
	  tl=tl->next;
	}
      if (bn==d->end)
	break;
      bn=bn->next;
      putc((bn->state>>8)|0x80, out);
      putc((bn->state&0xFF), out);
      bc+=1;
    }
  fclose(out);
  return bc;
}

int data_debug(Block* d)
{
  BlockNode *bn=d->start;
  TextList *tl;
  int bc=1, x;
  char c;
/*   if (bn->prev!=NULL) */
/*     fprintf(stderr, "\"0x%04x\"",bn->prev->state); */
  fprintf(stderr,"\"0x%04x\"",(bn->state^bn->change));
  fprintf(stderr,"\"0x%04x\"",bn->state);
  while(1)
    {
      tl=bn->text;
      while(tl)
	{
	  //printf("%d ", tl->len);
	  for (x=0;x<(tl->len);x++)
	    {
	      c=tl->data[x];
	      if((c==0x09)||(c==0x0A)||(c==0x0D))
		c=' ';
	      putc(c,stderr);
	    }
/* 	      putc(tl->data[x],stderr); */
	  tl=tl->next;
	}
      if (bn==d->end)
	break;
      bn=bn->next;
      fprintf(stderr,"\"0x%04x\"",bn->state);
      /*       putc((bn->state>>8)|0x80, out); */
/*       putc((bn->state&0xFF), out); */
 bc+=1;
    }
      fprintf(stderr, "\n");
      return bc;
}

// Text list manipulation routines
void tl_prepend(BlockNode *bn, TextList *tl)
{
  TextList *temp;
  temp=bn->text;
  bn->text=tl;
  tl->next=temp;
}

void tl_append(BlockNode *bn, TextList *tl)
{
  TextList *temp;
  temp=bn->text_end;
  bn->text_end=tl;
  temp->next=tl;
  tl->next=NULL;
}

// Block length calculation
int block_length(Block* bl)
{
  BlockNode *c;
  int count=1;
  if (bl->start == bl->end)
    return 1;	// termination case
  c=bl->start;
  while(1)
    {
      c=c->next;
      count+=1;
      if (c == bl->end)
	return count;
    }
}

// Block insertion (Inserts a block node before the specified block node)
void block_insert(BlockNode* c, short state)
{
  BlockNode *n=BN_Alloc();
  n->text=TL_Alloc(0);
  n->text->len=0;
  n->text->next=NULL;
  n->text_end=n->text;
  n->state=state;
  // now for the pointers
  n->prev=c->prev;
  n->next=c;
  c->prev=n;
  if (n->prev != NULL)
    n->prev->next=n;
  fix_change(n);
}
