#include <stdio.h>
#include "sml_tags.h"

int spaces=0;
FILE* out;
short gstate=0, sstate=0,t;
char c;

short get_sstate(short s)
{
  short t;
  t=s&0x07FF;
  if (!(s&M_U))
    t=8|(t&0x07F0);
  return t;
}

int char_write(char d, short state)
{
  if (((d==' ')||(d==0x0D)||(d==0x0A)||(d==0x09))&&
      (!((state^gstate)&(M_SIZE|M_TT|M_U))) &&		// size, UL, TT are same
      ((!(M_U&state))|| (!((state^gstate)&M_COL))))
    {	// we can move whitespace ahead of some tags
      //fprintf(stderr, "Moving whitespace\n");
      putc(d, out);
      //if (state!=gstate)
	//{
	  //putc(0x80|(state>>8), out);
	  //putc(state & 0xFF, out);
	  //gstate=state;
	//}
      return 0;
    }
  if (state!=gstate)
    {
      putc(0x80|(state>>8), out);
      putc(state & 0xFF, out);
      gstate=state;
    }
  putc(d, out);
  return 0;
}

short tag_decode(char d)
{
  switch(d)
    {
    case 'B':
      return M_B;
    case 'E':
      return M_EM;
    case 'I':
      return M_I;
    case 'P':
      return M_PL;
    case 'S':
      return M_S;
    case 'T':
      return M_TT;
    case 'U':
      return M_U;
    case 'r':
      return 1;
    case 'g':
      return 2;
    case 'b':
      return 3;
    case 'c':
      return 4;
    case 'm':
      return 5;
    case 'y':
      return 6;
    case 'k':
      return 7;
    case 'w':
      return 8;
    default:
      if ((d>='0')&&(d<='9'))
	return ((d+1-'0')<<4);
      return 0;
    }
}


int parse(short state, short prev)
{
  //Recursive function to read and canonicalize the input file
  //char c;
  //short t;
  //fprintf(stderr, "New parse...\n");
  while (1)
    {
      c=getchar();
      if (c=='<')	//beginning of a tag
        {
	  c=getchar();
	  if (c=='/')	//closing tag
	    {
	      t=tag_decode(getchar());
	      while (c!='>')
		{
		  c=getchar();
		  if (feof(stdin))
		    return 2;
		}
	      if (t==prev)
		return 0;
	      //Uh oh, tags aren't matching.  Error.
	      fprintf(stderr, "Error: Non-matching tags found\n");
	      exit(2);
	    }
	  //Opening tag
	  t=tag_decode(c);
	  //fprintf(stderr,"Got tag %04x from %c\n", t, c);
	  while (c!='>')
	    {
	      c=getchar();
	      if (feof(stdin))
		return 2;
	    }
	  if (t & M_COL)	// Color tag
	    {
/* 	      if (state & M_U) */
/* 		spaces=0; */
	      t=parse((state & ~M_COL)|t, t);
/* 	      if (state & M_U) */
/* 		spaces=0; */
	    }
	  else if (t & M_SIZE)	// Size
	    {
/* 	      spaces=0; */
	      t=parse((state& ~M_SIZE)|t, t);
/* 	      spaces=0; */
	    }
	  else if (t & M_U)	// Underline
	    {
	      if ((state&M_U)==M_U)
		t=parse(state, t);
	      else
		{
/* 		  spaces=0; */
		  t=parse(state+0x100, t);
/* 		  spaces=0; */
		}
	    }
	  else if (t & M_EM)	// Emphasis
	    {
	      if (state & M_S)
		t=parse(state, t);
	      else
		t=parse(state^M_EM, t);
	    }
	  else if (t & M_PL)	// PL
	    t=parse(state & ~(M_S|M_EM|M_I|M_B|M_TT|M_U), t);
	  else if (t & M_S)	// S
	    t=parse((state & ~M_EM)|M_S, t);
	  else if (t)
	    t=parse(state|t, t);
	  else
	    {
	      fprintf(stderr, "Unrecognized tag!\n");
	      exit(2);
	    }
	  //fprintf(stderr, "...Completed\n");
	  if (t!=0)
	    return t;
	}
      else if (c=='>')	//end of tag, should not occur!
        {
	  fprintf(stderr, "Found \">\", but not as part of a tag\n");
	  exit(1);
	}
      else if ((c==' ')||(c==0x0D)||(c==0x0A)||(c==0x09))
        {		//Whitespace
	  if ((spaces==0) || (state & M_TT) || (sstate!=get_sstate(state)))
	    {
	      if (state & M_TT)
		char_write(c, state);
	      else
		char_write(' ', state);
	    }
	  //else
	  //fprintf(stderr,"Omitting space\n");
	  if (state & M_TT) /* ||(sstate!=get_sstate(state))) */
	    { spaces=0; sstate=get_sstate(state); }
	  else if (sstate!=get_sstate(state))
	    { spaces=1; sstate=get_sstate(state); }
	  else
	    spaces++;
	}
      else if ((c>' ')&&(c<0x7F))
        {		//Other characters
	  char_write(c, state);
	  spaces=0;
	}
      else		//unknown character or EOF
        {
	  if (feof(stdin))
	    return 1;	//normal exit
	  fprintf(stderr,"Unknown character 0x%2x found in input\n", 0xFF&c);
	  return 2;
	}
    }
}

void usage()
{
  fprintf(stderr, "Need an output file...\n");
  exit (1);
}

int main(int argc, char** argv)
{
  //Assume that the input file is on stdin and we are writing
  // to the file in argv[1]
  
  //End declarations, start working...
  if (argc==1)
    usage();
  out=fopen(argv[1], "w");
  if (out==NULL)
    {
      fprintf(stderr, "Cannot open the output file %s\n", argv[1]);
      usage();
    }
  // Call the parser...
  spaces=parse(0,0);
  if (gstate)
    {
      putc(0x80, out);
      putc(0x00, out);
    }
  return spaces;
}
